package com.trilead.ssh2.transport;

/**
 * NegotiateException.
 * 
 * @author Christian Plattner, plattner@trilead.com
 * @version $Id: NegotiateException.java,v 1.1 2007/10/15 12:49:56 cplattne Exp $
 */
public class NegotiateException extends Exception
{
	private static final long serialVersionUID = 3689910669428143157L;
}
